/**
 * @file mmc.c
 *
 * Implementation of SD/MMC protocol
 * Based on Linux and U-Boot source code
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "mmc.h"
#include <stdio.h>
#include <rtems/endian.h>

//#define MMC_DEBUG(fmt, ...) printk("%s:" fmt, __func__, ##__VA_ARGS__)
#define MMC_DEBUG(fmt, ...)

//
// Private Services
//

static uint32_t mmc_get_voltages(struct mmc *mmc)
{
	if (mmc->device->get_voltages == NULL)
		return 0;

	return mmc->device->get_voltages();
}

static uint32_t mmc_is_spi(struct mmc *mmc)
{
	if (mmc->device->is_spi == NULL)
		return 0;

	return mmc->device->is_spi();
}

static rtems_status_code mmc_set_ios(struct mmc *mmc)
{
	if (mmc->device->set_ios == NULL)
		return RTEMS_NOT_IMPLEMENTED;

	return mmc->device->set_ios(&mmc->ios);
}

static rtems_status_code mmc_send_command(struct mmc *mmc, struct mmc_command *cmd)
{
	if (mmc->device->send_command == NULL)
		return RTEMS_NOT_IMPLEMENTED;

	return mmc->device->send_command(cmd);
}

static int mmc_card_is_sd(struct mmc *mmc)
{
	return mmc->version == MMC_VERSION_SD_1_0
			|| mmc->version == MMC_VERSION_SD_1_10
			|| mmc->version == MMC_VERSION_SD_2;
}

static rtems_status_code mmc_go_idle(struct mmc *mmc)
{
	struct mmc_command cmd;
	rtems_status_code err;

	cmd.code = MMC_CMD_GO_IDLE_STATE;
	cmd.arg = 0;
	cmd.data = NULL;
	cmd.resp_type = MMC_CMD_RESP_0;
	err = mmc_send_command(mmc, &cmd);

	if (err)
		return err;

	// wait for card go to idle
	rtems_task_wake_after(2 * 1000 / rtems_clock_get_ticks_per_second());

	return RTEMS_SUCCESSFUL;
}

static rtems_status_code mmc_send_if_cond(struct mmc *mmc)
{
	struct mmc_command cmd;
	int err;

	cmd.code = MMC_CMD_SEND_IF_COND;
	/* We set the bit if the host supports voltages between 2.7 and 3.6 V */
	cmd.arg = ((mmc_get_voltages(mmc) & 0xff8000) != 0) << 8 | 0xaa;
	cmd.data = NULL;
	cmd.resp_type = MMC_CMD_RESP_3;
	err = mmc_send_command(mmc, &cmd);

	if (err != RTEMS_SUCCESSFUL)
		return err;

	if ((cmd.resp[0] & 0xff) != 0xaa)
		return RTEMS_UNSATISFIED;

	mmc->version = MMC_VERSION_SD_2;
	return RTEMS_SUCCESSFUL;
}

static rtems_status_code mmc_sd_send_op_cond(struct mmc *mmc)
{
	struct mmc_command cmd;
	int timeout = 100;
	int err;

	do {
		cmd.code = MMC_CMD_APP_CMD;
		cmd.arg = 0;
		cmd.data = NULL;
		cmd.resp_type = MMC_CMD_RESP_3;
		err = mmc_send_command(mmc, &cmd);
		if (err)
			return err;

		cmd.code = MMC_CMD_APP_SEND_OP_COND;
		cmd.arg = mmc_is_spi(mmc) ? 0 : (mmc_get_voltages(mmc) & 0xff8000);
		if (mmc->version == MMC_VERSION_SD_2)
			cmd.arg |= MMC_OCR_HCS;
		cmd.data = NULL;
		cmd.resp_type = MMC_CMD_RESP_3;

		err = mmc_send_command(mmc, &cmd);
		if (err)
			return err;

		rtems_task_wake_after(1 * rtems_clock_get_ticks_per_second() / 1000);
	} while ((!(cmd.resp[0] & MMC_OCR_BUSY)) && timeout--);

	if (timeout <= 0)
		return RTEMS_TIMEOUT;

	if (mmc->version != MMC_VERSION_SD_2)
		mmc->version = MMC_VERSION_SD_1_0;

	mmc->ocr = cmd.resp[0];

	mmc->high_capacity = ((mmc->ocr & MMC_OCR_HCS) == MMC_OCR_HCS);
	mmc->rca = 0;

	return RTEMS_SUCCESSFUL;
}

static rtems_status_code mmc_send_status(struct mmc *mmc, int timeout)
{
	struct mmc_command cmd;
	int err, retries = 5;

	cmd.code = MMC_CMD_SEND_STATUS;
	cmd.resp_type = MMC_CMD_RESP_1;
	if (!mmc_is_spi(mmc))
		cmd.arg = mmc->rca << 16;
	cmd.data = NULL;

	do {
		err = mmc_send_command(mmc, &cmd);
		if (err == RTEMS_SUCCESSFUL) {
			if ((cmd.resp[0] & MMC_STATUS_RDY_FOR_DATA) &&
			    (cmd.resp[0] & MMC_STATUS_CURR_STATE) != MMC_STATE_PRG)
			{
				break;
			}
			else if (cmd.resp[0] & MMC_STATUS_MASK)
			{
				return RTEMS_IO_ERROR;
			}
		} else if (--retries < 0)
			return err;

		rtems_task_wake_after(1 * rtems_clock_get_ticks_per_second() / 1000);

	} while (timeout--);

	if (timeout <= 0) {
		return RTEMS_TIMEOUT;
	}

	return RTEMS_SUCCESSFUL;
}
/*
static int mmc_sd_switch(struct mmc *mmc, int mode, int group, uint8_t value, uint8_t *resp)
{
	struct mmc_command cmd;
	struct mmc_data data;

	// Switch the frequency
	cmd.code = MMC_SD_CMD_SWITCH_FUNC;
	cmd.resp_type = MMC_CMD_RESP_1;
	cmd.arg = (mode << 31) | 0xffffff;
	cmd.arg &= ~(0xf << (group * 4));
	cmd.arg |= value << (group * 4);
	cmd.data = &data;

	data.buffer = resp;
	data.blocksize = 64;
	data.blocks = 1;
	data.flags = MMC_DATA_READ;

	return mmc_send_command(mmc, &cmd);
}
*/

static rtems_status_code mmc_sd_change_freq(struct mmc *mmc)
{
	int err;
	struct mmc_command cmd;
	struct mmc_data data;
	int timeout;
	static uint32_t scr[2];

	mmc->card_caps = 0;

	if (mmc_is_spi(mmc))
		return RTEMS_SUCCESSFUL;

	/* Read the SCR to find out if this card supports higher speeds */
	cmd.code = MMC_CMD_APP_CMD;
	cmd.arg = mmc->rca << 16;
	cmd.data = NULL;
	cmd.resp_type = MMC_CMD_RESP_1;

	err = mmc_send_command(mmc, &cmd);

	if (err)
		return err;

	cmd.code = MMC_CMD_APP_SEND_SCR;
	cmd.resp_type = MMC_CMD_RESP_1;
	cmd.arg = 0;
	cmd.data = &data;

	timeout = 3;

retry_scr:
	data.buffer = (uint8_t*) scr;
	data.blocksize = 8;
	data.blocks = 1;
	data.flags = MMC_DATA_READ;

	err = mmc_send_command(mmc, &cmd);

	if (err) {
		if (timeout--)
			goto retry_scr;

		return err;
	}

	mmc->scr[0] = rtems_uint32_from_big_endian((uint8_t *) &scr[0]);
	mmc->scr[1] = rtems_uint32_from_big_endian((uint8_t *) &scr[1]);

	switch ((mmc->scr[0] >> 24) & 0xf) {
		case 0:
			mmc->version = MMC_VERSION_SD_1_0;
			break;
		case 1:
			mmc->version = MMC_VERSION_SD_1_10;
			break;
		case 2:
			mmc->version = MMC_VERSION_SD_2;
			break;
		default:
			mmc->version = MMC_VERSION_SD_1_0;
			break;
	}

	if (mmc->scr[0] & SD_DATA_4BIT)
		mmc->card_caps |= MMC_MODE_4BIT;

	return RTEMS_SUCCESSFUL;
}

/* frequency bases */
/* divided by 10 to be nice to platforms without floating point */
static const int fbase[] = {
	10000,
	100000,
	1000000,
	10000000,
};

/* Multiplier values for TRAN_SPEED.  Multiplied by 10 to be nice
 * to platforms without floating point.
 */
static const int multipliers[] = {
	0,	/* reserved */
	10,
	12,
	13,
	15,
	20,
	25,
	30,
	35,
	40,
	45,
	50,
	55,
	60,
	70,
	80,
};

static rtems_status_code mmc_startup(struct mmc *mmc)
{
	struct mmc_command cmd;
	int err;
	uint32_t mult, freq;
	uint64_t cmult, csize;

	/* Put the Card in Identify Mode */
	cmd.code = mmc_is_spi(mmc) ? MMC_CMD_SEND_CID : MMC_CMD_ALL_SEND_CID; /* cmd not supported in spi */
	cmd.resp_type = MMC_CMD_RESP_2;
	cmd.arg = 0;
	cmd.data = NULL;

	err = mmc_send_command(mmc, &cmd);

	if (err)
		return err;

	mmc->cid[0] = cmd.resp[0];
	mmc->cid[1] = cmd.resp[1];
	mmc->cid[2] = cmd.resp[2];
	mmc->cid[3] = cmd.resp[3];

	/*
	 * For MMC cards, set the Relative Address.
	 * For SD cards, get the Relative Address.
	 * This also puts the cards into Standby State
	 */
	if (!mmc_is_spi(mmc)) { /* cmd not supported in spi */
		cmd.code = MMC_CMD_SEND_RELATIVE_ADDR;
		cmd.arg = mmc->rca << 16;
		cmd.data = NULL;
		cmd.resp_type = MMC_CMD_RESP_3;

		err = mmc_send_command(mmc, &cmd);

		if (err)
			return err;

		if (mmc_card_is_sd(mmc))
			mmc->rca = (cmd.resp[0] >> 16) & 0xffff;
	}

	/* Get the Card-Specific Data */
	cmd.code = MMC_CMD_SEND_CSD;
	cmd.resp_type = MMC_CMD_RESP_2;
	cmd.arg = mmc->rca << 16;
	cmd.data = NULL;

	err = mmc_send_command(mmc, &cmd);

	/* Waiting for the ready status */
	mmc_send_status(mmc, 1000);

	if (err)
		return err;

	mmc->csd[0] = cmd.resp[0];
	mmc->csd[1] = cmd.resp[1];
	mmc->csd[2] = cmd.resp[2];
	mmc->csd[3] = cmd.resp[3];

	if (mmc->version == MMC_VERSION_UNKNOWN) {
		int version = (cmd.resp[0] >> 26) & 0xf;

		switch (version) {
			case 0:
				mmc->version = MMC_VERSION_MMC_1_2;
				break;
			case 1:
				mmc->version = MMC_VERSION_MMC_1_4;
				break;
			case 2:
				mmc->version = MMC_VERSION_MMC_2_2;
				break;
			case 3:
				mmc->version = MMC_VERSION_MMC_3;
				break;
			case 4:
				mmc->version = MMC_VERSION_MMC_4;
				break;
			default:
				mmc->version = MMC_VERSION_MMC_1_2;
				break;
		}
	}

	/* divide frequency by 10, since the mults are 10x bigger */
	freq = fbase[(cmd.resp[0] & 0x7)];
	mult = multipliers[((cmd.resp[0] >> 3) & 0xf)];

	mmc->tran_speed = freq * mult;
	mmc->read_bl_len = 1 << ((cmd.resp[1] >> 16) & 0xf);

	if (mmc_card_is_sd(mmc))
		mmc->write_bl_len = mmc->read_bl_len;
	else
		mmc->write_bl_len = 1 << ((cmd.resp[3] >> 22) & 0xf);

	if (mmc->high_capacity)
	{
		csize = (mmc->csd[1] & 0x3f) << 16 | (mmc->csd[2] & 0xffff0000) >> 16;
		cmult = 8;
	}
	else
	{
		csize = (mmc->csd[1] & 0x3ff) << 2 | (mmc->csd[2] & 0xc0000000) >> 30;
		cmult = (mmc->csd[2] & 0x00038000) >> 15;
	}

	mmc->capacity = (csize + 1) << (cmult + 2);
	mmc->capacity *= mmc->read_bl_len;

	if (mmc->read_bl_len > 512)
		mmc->read_bl_len = 512;

	if (mmc->write_bl_len > 512)
		mmc->write_bl_len = 512;

	mmc->blocks_count = mmc->capacity / mmc->read_bl_len;

	/* Select the card, and put it into Transfer Mode */
	if (!mmc_is_spi(mmc)) { /* cmd not supported in spi */
		cmd.code = MMC_CMD_SELECT_CARD;
		cmd.resp_type = MMC_CMD_RESP_1;
		cmd.arg = mmc->rca << 16;
		cmd.data = NULL;
		err = mmc_send_command(mmc, &cmd);

		if (err)
			return err;
	}

	mmc->erase_grp_size = 1;

	MMC_DEBUG("Found SDCard: blocks=%lu read_bl_len=%lu write_bl_len=%lu erase_grp_size=%lu\n",
			mmc->blocks_count, mmc->read_bl_len, mmc->write_bl_len, mmc->erase_grp_size);

	if (mmc_card_is_sd(mmc))
		err = mmc_sd_change_freq(mmc);
	else
		return RTEMS_NOT_IMPLEMENTED;

	if (err != RTEMS_SUCCESSFUL)
		return err;

	if (mmc_card_is_sd(mmc))
	{
		if (mmc->card_caps & MMC_MODE_4BIT)
		{
			cmd.code = MMC_CMD_APP_CMD;
			cmd.resp_type = MMC_CMD_RESP_1;
			cmd.arg = mmc->rca << 16;
			cmd.data = NULL;
			err = mmc_send_command(mmc, &cmd);
			if (err)
				return err;

			cmd.code = MMC_CMD_APP_SET_BUS_WIDTH;
			cmd.resp_type = MMC_CMD_RESP_1;
			cmd.arg = 2;
			cmd.data = NULL;
			err = mmc_send_command(mmc, &cmd);
			if (err)
				return err;

			mmc->ios.bus_width = 4;
		}
		mmc->ios.clock = mmc->tran_speed = 25000000;
	}
	else
	{
		return RTEMS_NOT_IMPLEMENTED;
	}

	mmc_set_ios(mmc);

	return RTEMS_SUCCESSFUL;
}

static rtems_status_code mmc_set_blocklen(struct mmc *mmc, uint32_t blocklen)
{
	struct mmc_command cmd;

	MMC_DEBUG("Send 'Set block length' to %d\n", blocklen);

	cmd.code = MMC_CMD_SET_BLOCKLEN;
	cmd.resp_type = MMC_CMD_RESP_1;
	cmd.arg = blocklen;
	cmd.data = NULL;

	return mmc_send_command(mmc, &cmd);
}

static rtems_status_code mmc_read_blocks(struct mmc *mmc, uint32_t start, uint32_t count,
		void *buffer, size_t size)
{
	struct mmc_command cmd;
	struct mmc_data data;
	rtems_status_code err;

	MMC_DEBUG("Read blocks from %lu count %lu\n", start, count);

	if (count > 1)
		cmd.code = MMC_CMD_READ_MULTIPLE_BLOCK;
	else
		cmd.code = MMC_CMD_READ_SINGLE_BLOCK;

	if (mmc->high_capacity)
		cmd.arg = start;
	else
		cmd.arg = start * mmc->read_bl_len;

	cmd.data = &data;
	cmd.resp_type = MMC_CMD_RESP_1;

	data.buffer = buffer;
	data.blocks = count;
	data.blocksize = mmc->read_bl_len;
	data.flags = MMC_DATA_READ;

	err = mmc_send_command(mmc, &cmd);
	if (err != RTEMS_SUCCESSFUL)
		return err;

	if (count > 1)
	{
		cmd.code = MMC_CMD_STOP_TRANSMISSION;
		cmd.arg = 0;
		cmd.data = NULL;
		cmd.resp_type = MMC_CMD_RESP_1;
		err = mmc_send_command(mmc, &cmd);
		if (err != RTEMS_SUCCESSFUL)
		{
			printk("MMC: fail to send stop cmd\n");
			return err;
		}
	}

	return RTEMS_SUCCESSFUL;
}

static rtems_status_code mmc_write_blocks(struct mmc *mmc, uint32_t start, uint32_t blkcnt, const void *buffer, size_t size)
{
	struct mmc_command cmd;
	struct mmc_data data;
	int timeout = 1000;
	rtems_status_code err;

	MMC_DEBUG("Write blocks from %lu count %lu\n", start, blkcnt);

	if ((start + blkcnt) > mmc->blocks_count) {
		printk("MMC: block number 0x%lx exceeds max(0x%lx)\n",
			start + blkcnt, mmc->blocks_count);
		return RTEMS_INVALID_ADDRESS;
	}

	if (blkcnt > 1)
		cmd.code = MMC_CMD_WRITE_MULTIPLE_BLOCK;
	else
		cmd.code = MMC_CMD_WRITE_SINGLE_BLOCK;

	if (mmc->high_capacity)
		cmd.arg = start;
	else
		cmd.arg = start * mmc->write_bl_len;

	cmd.data = &data;
	cmd.resp_type = MMC_CMD_RESP_1;

	data.buffer = (void *) buffer;
	data.blocks = blkcnt;
	data.blocksize = mmc->write_bl_len;
	data.flags = MMC_DATA_WRITE;

	err = mmc_send_command(mmc, &cmd);
	if (err != RTEMS_SUCCESSFUL)
		return err;

	/* SPI multiblock writes terminate using a special
	 * token, not a STOP_TRANSMISSION request.
	 */
	if (!mmc_is_spi(mmc) && blkcnt > 1) {
		cmd.code = MMC_CMD_STOP_TRANSMISSION;
		cmd.arg = 0;
		cmd.data = NULL;
		cmd.resp_type = MMC_CMD_RESP_1;
		err = mmc_send_command(mmc, &cmd);
		if (err != RTEMS_SUCCESSFUL)
		{
			printk("MMC: fail to send stop cmd\n");
			return err;
		}
	}

	/* Waiting for the ready status */
	err = mmc_send_status(mmc, timeout);
	if (err != RTEMS_SUCCESSFUL)
		return err;

	return RTEMS_SUCCESSFUL;
}


static uint32_t mmc_erase_t(struct mmc *mmc, uint32_t start, uint32_t blkcnt)
{
	struct mmc_command cmd;
	uint32_t end;
	int err, start_cmd, end_cmd;

	if (mmc->high_capacity)
		end = start + blkcnt - 1;
	else {
		end = (start + blkcnt - 1) * mmc->write_bl_len;
		start *= mmc->write_bl_len;
	}

	if (mmc_card_is_sd(mmc)) {
		start_cmd = MMC_SD_CMD_ERASE_WR_BLK_START;
		end_cmd = MMC_SD_CMD_ERASE_WR_BLK_END;
	} else {
		start_cmd = MMC_CMD_ERASE_GROUP_START;
		end_cmd = MMC_CMD_ERASE_GROUP_END;
	}

	cmd.code = start_cmd;
	cmd.arg = start;
	cmd.data = NULL;
	cmd.resp_type = MMC_CMD_RESP_1;

	err = mmc_send_command(mmc, &cmd);
	if (err)
		goto err_out;

	cmd.code = end_cmd;
	cmd.arg = end;
	cmd.data = NULL;
	cmd.resp_type = MMC_CMD_RESP_1;

	err = mmc_send_command(mmc, &cmd);
	if (err)
		goto err_out;

	cmd.code = MMC_CMD_ERASE;
	cmd.arg = MMC_SECURE_ERASE;
	cmd.data = NULL;
	cmd.resp_type = MMC_CMD_RESP_1;

	err = mmc_send_command(mmc, &cmd);
	if (err)
		goto err_out;

	return 0;

err_out:
	printk("mmc erase failed\n");
	return err;
}

//
// Public Services
//

rtems_status_code mmc_init(struct mmc *mmc, struct mmc_device *device)
{
	int err;

	memset(mmc, 0, sizeof(*mmc));
	mmc->device = device;

	if (mmc->device->probe)
	{
		err = mmc->device->probe();
		if (err != RTEMS_SUCCESSFUL)
			return err;
	}

	if (mmc->device->card_detect && mmc->device->card_detect() == 0)
		return RTEMS_IO_ERROR;

	MMC_DEBUG("Send 'Go Idle'\n");
	mmc->ios.bus_width = 1;
	mmc->ios.clock = 400 * 1000; // Clock frequency Identification Mode (OD)
	mmc_set_ios(mmc);

	/* Reset the Card */
	err = mmc_go_idle(mmc);

	if (err)
		return err;

	MMC_DEBUG("Send 'If Cond'\n");

	/* Test for SD version 2 */
	mmc_send_if_cond(mmc);

	/* Now try to get the SD card's operating condition */
	err = mmc_sd_send_op_cond(mmc);

	if (err == RTEMS_TIMEOUT) {
		/* If the command timed out, we check for an MMC card */
		return RTEMS_NOT_IMPLEMENTED;
	}

	MMC_DEBUG("MMC startup\n");

	return mmc_startup(mmc);
}

rtems_status_code mmc_free(struct mmc *mmc)
{
	return RTEMS_SUCCESSFUL;
}

uint32_t mmc_block_len(struct mmc *mmc)
{
	return mmc->read_bl_len;
}

uint32_t mmc_block_count(struct mmc *mmc)
{
	return mmc->blocks_count;
}

rtems_status_code mmc_read(struct mmc *mmc,  uint32_t start, uint32_t count, void *buffer, size_t size)
{
	rtems_status_code sc;

	if (start + count >= mmc->blocks_count)
		return RTEMS_INVALID_ADDRESS;

	if (mmc->read_bl_len * count != size)
		return RTEMS_INVALID_NUMBER;

	sc = mmc_set_blocklen(mmc, mmc->read_bl_len);
	if (sc == RTEMS_SUCCESSFUL)
		sc = mmc_read_blocks(mmc, start, count, buffer, size);

	return sc;
}

rtems_status_code mmc_write(struct mmc *mmc,  uint32_t start, uint32_t count, const void *buffer, size_t size)
{
	rtems_status_code sc;

	sc = mmc_set_blocklen(mmc, mmc->write_bl_len);
	if (sc == RTEMS_SUCCESSFUL)
		sc = mmc_write_blocks(mmc, start, count, buffer, size);

	return sc;
}

rtems_status_code mmc_erase(struct mmc *mmc,  uint32_t start, uint32_t count)
{
	uint32_t cur = 0;
	uint32_t cur_count;
	rtems_status_code err = RTEMS_SUCCESSFUL;

	if ((start % mmc->erase_grp_size) || (count % mmc->erase_grp_size))
	{
		printk("Can't erase blocks from %lu to %lu, erase group is %lu\n",
				start, count, mmc->erase_grp_size);
		return RTEMS_INVALID_NUMBER;
	}

	while (cur < count) {
		cur_count = ((count - cur) > mmc->erase_grp_size) ?
			mmc->erase_grp_size : (count - cur);
		err = mmc_erase_t(mmc, start + cur, cur_count);
		if (err != RTEMS_SUCCESSFUL)
			break;

		cur += cur_count;

		/* Waiting for the ready status */
		err = mmc_send_status(mmc, 1000);
		if (err != RTEMS_SUCCESSFUL)
			break;
	}

	return err;
}
