/**
 * @file mmc.h
 *
 * Implementation of SD/MMC protocol
 * Based on Linux and U-Boot source code
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef MMC_H_
#define MMC_H_

#include <stdint.h>
#include <rtems.h>

//
// Public Types
//

struct mmc_ios
{
	int bus_width;		///< count of wire in bus
	int clock;			///< clock rate in Hz
};

/* MMC command codes */
#define MMC_CMD_GO_IDLE_STATE			0
#define MMC_CMD_SEND_OP_COND			1
#define MMC_CMD_ALL_SEND_CID			2
#define MMC_CMD_SEND_RELATIVE_ADDR		3
#define MMC_CMD_APP_SET_BUS_WIDTH		6
#define MMC_CMD_SELECT_CARD				7
#define MMC_CMD_SEND_IF_COND			8
#define MMC_CMD_SEND_CSD				9
#define MMC_CMD_SEND_CID				10
#define MMC_CMD_STOP_TRANSMISSION		12
#define MMC_CMD_SEND_STATUS				13
#define MMC_CMD_SET_BLOCKLEN			16
#define MMC_CMD_READ_SINGLE_BLOCK		17
#define MMC_CMD_READ_MULTIPLE_BLOCK		18
#define MMC_CMD_WRITE_SINGLE_BLOCK		24
#define MMC_CMD_WRITE_MULTIPLE_BLOCK	25
#define MMC_CMD_ERASE_GROUP_START		35
#define MMC_CMD_ERASE_GROUP_END			36
#define MMC_CMD_ERASE					38

#define MMC_CMD_APP_SEND_OP_COND		41
#define MMC_CMD_APP_SEND_SCR			51
#define MMC_CMD_APP_CMD					55
#define MMC_SD_CMD_SWITCH_FUNC			6
#define MMC_SD_CMD_ERASE_WR_BLK_START	32
#define MMC_SD_CMD_ERASE_WR_BLK_END		33

#define MMC_CMD_RESP_0					0
#define MMC_CMD_RESP_1					1
#define MMC_CMD_RESP_2					2
#define MMC_CMD_RESP_3					3

#define MMC_OCR_HCS						0x40000000
#define MMC_OCR_BUSY					0x80000000

#define MMC_MODE_HS						0x001
#define MMC_MODE_HS_52MHz				0x010
#define MMC_MODE_4BIT					0x100
#define MMC_MODE_8BIT					0x200
#define MMC_MODE_SPI					0x400
#define MMC_MODE_HC						0x800

#define MMC_MODE_MASK_WIDTH_BITS (MMC_MODE_4BIT | MMC_MODE_8BIT)
#define MMC_MODE_WIDTH_BITS_SHIFT 8

#define SD_DATA_4BIT					0x00040000

#define SD_SWITCH_CHECK					0
#define SD_SWITCH_SWITCH				1

#define MMC_SECURE_ERASE				0x80000000

/* SCR definitions in different words */
#define SD_HIGHSPEED_BUSY	0x00020000
#define SD_HIGHSPEED_SUPPORTED	0x00020000

#define MMC_STATUS_MASK					(~0x0206BF7F)
#define MMC_STATUS_RDY_FOR_DATA			(1 << 8)
#define MMC_STATUS_CURR_STATE			(0xf << 9)
#define MMC_STATUS_ERROR				(1 << 19)
#define MMC_STATE_PRG					(7 << 9)

#define MMC_DATA_READ					1
#define MMC_DATA_WRITE					2

struct mmc_data
{
	uint8_t *buffer;
	uint32_t flags;
	uint32_t blocks;
	uint32_t blocksize;
};

struct mmc_command
{
	uint32_t		code;
	uint32_t 		arg;
	struct mmc_data *data;
	uint32_t		resp_type;
	uint32_t 		resp[4];
};

struct mmc_device {
	/**
	 * Probe device and initialization
	 * @return status code
	 */
	rtems_status_code (*probe)(void);

	/**
	 * Check for card available
	 * @return Zero then not available
	 */
	int (*card_detect)(void);

	/**
	 * Get type of physical connection
	 * @return Non-zero if SPI mode
	 */
	int (*is_spi)(void);

	/**
	 * Return supported voltages by device
	 * @return Voltages range in OCR register format
	 */
	uint32_t (*get_voltages)(void);

	/**
	 * Setup IO
	 * @param ios Configuration
	 * @return status code
	 */
	rtems_status_code (*set_ios)(struct mmc_ios *ios);

	/**
	 * Send MMC command
	 * @param command Command structure
	 * @return status code
	 */
	rtems_status_code (*send_command)(struct mmc_command *command);
};

enum mmc_version
{
	MMC_VERSION_UNKNOWN	= 0,
	MMC_VERSION_SD_1_0,
	MMC_VERSION_SD_1_10,
	MMC_VERSION_SD_2,
	MMC_VERSION_MMC_1_2,
	MMC_VERSION_MMC_1_4,
	MMC_VERSION_MMC_2_2,
	MMC_VERSION_MMC_3,
	MMC_VERSION_MMC_4,
};

struct mmc {
	struct mmc_device *device;
	struct mmc_ios ios;
	enum mmc_version version;
	uint32_t ocr;
	uint32_t cid[4];
	uint32_t csd[4];
	int	high_capacity;
	uint16_t rca;
	uint32_t card_caps;
	uint32_t tran_speed;
	uint32_t read_bl_len;
	uint32_t write_bl_len;
	uint32_t erase_grp_size;
	uint64_t capacity;
	uint32_t blocks_count;
	uint32_t scr[2];
};

//
// Public Services
//

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Initialize resources of MMC instance
 * @return status code
 */
rtems_status_code mmc_init(struct mmc *mmc, struct mmc_device *device);

/**
 * Free resources of MMC instance
 * @return status code
 */
rtems_status_code mmc_free(struct mmc *mmc);

/**
 * Get length of MMC device block
 * @param mmc MMC instance
 * @return length of block
 */
uint32_t mmc_block_len(struct mmc *mmc);

/**
 * Get count of MMC device blocks
 * @param mmc MMC instance
 * @return count of blocks
 */
uint32_t mmc_block_count(struct mmc *mmc);

/**
 * Read blocks from MMC
 * @param mmc MMC instance
 * @param start Start block number
 * @param count Count of blocks to read
 * @param buffer Destination buffer
 * @param size Size of destination buffer
 * @return status code
 */
rtems_status_code mmc_read(struct mmc *mmc,  uint32_t start, uint32_t count, void *buffer, size_t size);

/**
 * Write blocks to MMC
 * @param mmc MMC instance
 * @param start Start block number
 * @param count Count of blocks to write
 * @param buffer Source buffer
 * @param size Size of source buffer
 * @return status code
 */
rtems_status_code mmc_write(struct mmc *mmc,  uint32_t start, uint32_t count, const void *buffer, size_t size);

/**
 * Erase blocks on MMC
 * @param mmc MMC instance
 * @param start Start block number
 * @param count Count of blocks to erase
 * @return status code
 */
rtems_status_code mmc_erase(struct mmc *mmc,  uint32_t start, uint32_t count);

#ifdef __cplusplus
}
#endif

#endif /* MMC_H_ */
