/**
 * @file sdcard.c
 *
 * SD-Card block device driver for RTEMS
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <bsp/sdcard.h>
#include "mmc.h"
#include "mci.h"
#include <errno.h>
#include <rtems/error.h>

//
// Private Variables
//

static struct mmc mmc;
static rtems_id mutex;

//
// Private Services
//

static int sdcard_read(rtems_blkdev_request *req)
{
	uint32_t i;
	rtems_blkdev_sg_buffer *sg;
	rtems_status_code sc;
	int trycount;

	rtems_semaphore_obtain(mutex, RTEMS_WAIT, RTEMS_NO_TIMEOUT);

	for (i = 0, sg = req->bufs; i < req->bufnum; i++, sg++)
	{
		trycount = 10;
		do {
			sc = mmc_read(&mmc, sg->block, 1, sg->buffer, sg->length);
		} while (trycount > 10 && sc != RTEMS_SUCCESSFUL);
		if (sc != RTEMS_SUCCESSFUL)
			break;
	}

	if (sc != RTEMS_SUCCESSFUL)
    	printk ("sdcard read: start=%d, blocks=%d: %s\n",
    			req->bufs[0].block, req->bufnum, rtems_status_text(sc));

	req->status = sc;
	req->req_done(req->done_arg, sc);
    rtems_semaphore_release(mutex);

	return 0;
}

static int sdcard_write(rtems_blkdev_request *req)
{
	uint32_t i;
	rtems_blkdev_sg_buffer *sg;
	rtems_status_code sc;
	int trycount;

	rtems_semaphore_obtain(mutex, RTEMS_WAIT, RTEMS_NO_TIMEOUT);
    for (i = 0, sg = req->bufs; i < req->bufnum; i++, sg++)
    {
		trycount = 10;
		do {
			sc = mmc_erase(&mmc, sg->block, 1);
			if (sc == RTEMS_SUCCESSFUL)
				sc = mmc_write(&mmc, sg->block, 1, sg->buffer, sg->length);
		} while (trycount > 10 && sc != RTEMS_SUCCESSFUL);
		if (sc != RTEMS_SUCCESSFUL)
			break;
    }
    if (sc != RTEMS_SUCCESSFUL)
        printk("sdcard write: start=%d, blocks=%d: %s\n",
        		req->bufs[0].block, req->bufnum, rtems_status_text(sc));

    req->status = sc;
    req->req_done(req->done_arg, sc);
    rtems_semaphore_release(mutex);

    return 0;
}

static int sdcard_ioctl(rtems_disk_device *dd, uint32_t req, void *argp)
{
    switch (req)
    {
        case RTEMS_BLKIO_REQUEST:
        {
            rtems_blkdev_request *r = argp;

            switch (r->req)
            {
                case RTEMS_BLKDEV_REQ_READ:
                    return sdcard_read(r);

                case RTEMS_BLKDEV_REQ_WRITE:
                    return sdcard_write(r);

                default:
                    errno = EINVAL;
                    return -1;
            }
            break;
        }

        default:
            return rtems_blkdev_ioctl (dd, req, argp);
            break;
    }

    errno = EINVAL;
    return -1;
}


//
// Public Services
//

rtems_device_driver sdcard_initialize(rtems_device_major_number major, rtems_device_minor_number minor,
		void *arg)
{
	rtems_status_code rc;

	rc = rtems_disk_io_initialize();
	if (rc != RTEMS_SUCCESSFUL)
		return rc;

	rtems_semaphore_create(rtems_build_name('s', 'd', 'c', 'r'),	1,
				RTEMS_DEFAULT_ATTRIBUTES | RTEMS_BINARY_SEMAPHORE | RTEMS_INHERIT_PRIORITY | RTEMS_PRIORITY,
				RTEMS_NO_PRIORITY, &mutex);

	rc = mmc_init(&mmc, &mci_mmc_device);
	if (rc != RTEMS_SUCCESSFUL)
	{
		printk("mmc card init failed\n");
		return rc;
	}

	dev_t dev = rtems_filesystem_make_dev_t(major, 0);
    rc = rtems_disk_create_phys(dev, mmc_block_len(&mmc), mmc_block_count(&mmc),
    		sdcard_ioctl, NULL, "/dev/sda");
    if (rc != RTEMS_SUCCESSFUL)
    	return rc;

    return RTEMS_SUCCESSFUL;
}
